#! /usr/bin/env python2.7

"""Solve some tasks with A* and the LM-Cut heuristic."""

import os
import os.path
import platform

from lab.environments import LocalEnvironment, BaselSlurmEnvironment

from downward.experiment import FastDownwardExperiment
from downward.reports.absolute import AbsoluteReport
from downward.reports.scatter import ScatterPlotReport


ATTRIBUTES = ['coverage', 'error', 'expansions', 'total_time', 'plan_length', 'backwards_ops_in_plan', 'backwards_expansions', 'original_goal_found', 'initial_state_found', 'undone', 'goal_expansion_stopped']

NODE = platform.node()
if NODE.endswith(".scicore.unibas.ch") or NODE.endswith(".cluster.bc2.ch"):
    # Create bigger suites with suites.py from the downward-benchmarks repo.
    # This is optimal_strips
    SUITE = ['agricola-opt18-strips', 'airport', 'barman-opt11-strips', 'barman-opt14-strips', 'blocks', 'childsnack-opt14-strips', 'data-network-opt18-strips', 'depot', 'driverlog', 'elevators-opt08-strips', 'elevators-opt11-strips', 'floortile-opt11-strips', 'floortile-opt14-strips', 'freecell', 'ged-opt14-strips', 'grid', 'gripper', 'hiking-opt14-strips', 'logistics00', 'logistics98', 'miconic', 'movie', 'mprime', 'mystery', 'nomystery-opt11-strips', 'openstacks-opt08-strips', 'openstacks-opt11-strips', 'openstacks-opt14-strips', 'openstacks-strips', 'organic-synthesis-opt18-strips', 'organic-synthesis-split-opt18-strips', 'parcprinter-08-strips', 'parcprinter-opt11-strips', 'parking-opt11-strips', 'parking-opt14-strips', 'pathways-noneg', 'pegsol-08-strips', 'pegsol-opt11-strips', 'petri-net-alignment-opt18-strips', 'pipesworld-notankage', 'pipesworld-tankage', 'psr-small', 'rovers', 'satellite', 'scanalyzer-08-strips', 'scanalyzer-opt11-strips', 'snake-opt18-strips', 'sokoban-opt08-strips', 'sokoban-opt11-strips', 'spider-opt18-strips', 'storage', 'termes-opt18-strips', 'tetris-opt14-strips', 'tidybot-opt11-strips', 'tidybot-opt14-strips', 'tpp', 'transport-opt08-strips', 'transport-opt11-strips', 'transport-opt14-strips', 'trucks-strips', 'visitall-opt11-strips', 'visitall-opt14-strips', 'woodworking-opt08-strips', 'woodworking-opt11-strips', 'zenotravel']
    ENV = BaselSlurmEnvironment(email="remo.christen@stud.unibas.ch")
else:
    SUITE = ['depot:p01.pddl', 'gripper:prob01.pddl', 'no-mystery:prob01.pddl']
#    SUITE = ['gripper:prob01.pddl', 'gripper:prob02.pddl', 'gripper:prob03.pddl', 'gripper:prob04.pddl', 'gripper:prob05.pddl', 'gripper:prob06.pddl', 'gripper:prob07.pddl']
    ENV = LocalEnvironment(processes=2)
# Use path to your Fast Downward repository.
REPO = os.environ["DOWNWARD_REPO"]
BENCHMARKS_DIR = os.environ["DOWNWARD_BENCHMARKS"]
REVISION_CACHE = os.path.expanduser('~/lab/revision-cache')

exp = FastDownwardExperiment(environment=ENV, revision_cache=REVISION_CACHE)
sage_parser = os.path.abspath(os.path.join(__file__, '../sage_parser.py'))
exp.add_parser(sage_parser)
# Add built-in parsers to the experiment.
exp.add_parser(exp.EXITCODE_PARSER)
exp.add_parser(exp.TRANSLATOR_PARSER)
exp.add_parser(exp.SINGLE_SEARCH_PARSER)
exp.add_parser(exp.PLANNER_PARSER)
exp.add_suite(BENCHMARKS_DIR, SUITE)

exp.add_algorithm(
    'every time most sat',
    REPO,
    'expand_in_ff',
    ['--search', 'eager_greedy([ff(operator_order=most_satisfied_first)], use_bgg=true)'])

exp.add_algorithm(
    'every time lowest p',
    REPO,
    'expand_in_ff',
    ['--search', 'eager_greedy([ff(operator_order=lowest_proposition_layer_first)], use_bgg=true)'])

exp.add_algorithm(
    'new_min most sat',
    REPO,
    'expand_in_ff',
    ['--search', 'eager_greedy([ff(state_decision=new_min, operator_order=most_satisfied_first)], use_bgg=true)'])

exp.add_algorithm(
    'new_min lowest p',
    REPO,
    'expand_in_ff',
    ['--search', 'eager_greedy([ff(state_decision=new_min, operator_order=lowest_proposition_layer_first)], use_bgg=true)'])

exp.add_algorithm(
    'counter 25 80 most sat',
    REPO,
    'expand_in_ff',
    ['--search', 'eager_greedy([ff(state_decision=counter, state_value=25, percentage=80, operator_order=most_satisfied_first)], use_bgg=true)'])

exp.add_algorithm(
    'counter 25 80 lowest p',
    REPO,
    'expand_in_ff',
    ['--search', 'eager_greedy([ff(state_decision=counter, state_value=25, percentage=80, operator_order=lowest_proposition_layer_first)], use_bgg=true)'])

exp.add_algorithm(
    'diff 0 most sat',
    REPO,
    'expand_in_ff',
    ['--search', 'eager_greedy([ff(operator_order=most_satisfied_first, state_decision=diff, state_value=0)], use_bgg=true)'])

exp.add_algorithm(
    'diff 0 lowest p',
    REPO,
    'expand_in_ff',
    ['--search', 'eager_greedy([ff(operator_order=lowest_proposition_layer_first, state_decision=diff, state_value=0)], use_bgg=true)'])

#exp.add_algorithm(
#        'ff', REPO, 'default', ['--search', 'astar(ff())'], build_options=["--debug"], driver_options=["--debug"])

# Add step that writes experiment files to disk.
exp.add_step('build', exp.build)

# Add step that executes all runs.
exp.add_step('start', exp.start_runs)

# Add step that collects properties from run directories and
# writes them to *-eval/properties.
exp.add_fetcher(name='fetch')

# Add report step (AbsoluteReport is the standard report).
exp.add_report(
    AbsoluteReport(attributes=ATTRIBUTES), outfile='report.html')

# Add scatter plot report step.
exp.add_report(
    ScatterPlotReport(
        attributes=["expansions"], filter_algorithm=["blind", "lmcut"]),
    outfile='scatterplot.png')

# Parse the commandline and show or run experiment steps.
exp.run_steps()
